from a1facts.ontology.entity_class import EntityClass
from a1facts.ontology.property import Property
from a1facts.utils.logger import logger

class RelationshipClass:
    """Represents a class of relationships (edges) in the ontology."""
    def __init__(self, name: str, domain: EntityClass, range: EntityClass, description: str, symmetric: bool = False):
        """
        Initializes a RelationshipClass object.

        Args:
            name (str): The name of the relationship class (e.g., 'operates_in').
            domain (EntityClass): The domain (start node) entity class.
            range (EntityClass): The range (end node) entity class.
            description (str): A description of the relationship class.
            symmetric (bool): True if the relationship is symmetric.
        """
        self.relationship_name = name
        self.domain_entity_class = domain.entity_class_name
        self.domain_primary_key_prop = domain.primary_key_prop.property_name
        self.domain_primary_key_type = domain.primary_key_prop.type
        self.range_entity_class = range.entity_class_name
        self.range_primary_key_prop = range.primary_key_prop.property_name
        self.range_primary_key_type = range.primary_key_prop.type
        self.description = description
        self.properties = []
        self.symmetric = symmetric

    def add_property(self, property: Property):
        """
        Adds a property to the relationship class.

        Args:
            property (Property): The property to add.
        """
        self.properties.append(property)
    
    def __str__(self):
        """Returns a string representation of the relationship class."""
        relationship_str = ""
        relationship_str += f"{self.relationship_name} ({self.description}) - Domain: {self.domain_entity_class} - Range: {self.range_entity_class}\n"
        for prop in self.properties:
            relationship_str += f"   - {prop}\n"
        if self.symmetric:
            relationship_str += "   (This relationship is symmetric)\n"
        return relationship_str

    def is_symmetric(self):
        """Returns True if the relationship is symmetric."""
        return self.symmetric

    def _validate_properties(self, properties):
        """
        Validates that all required properties are present.

        Args:
            properties (dict): A dictionary of properties to validate.
        """
        if properties:
            for prop in self.properties:
                if prop.property_name not in properties:
                    raise Exception(f"Property {prop.property_name} not found in properties, you need to change the world model")

    def get_tool_add_or_update_relationship(self, add_or_update_relationship_func):
        """
        Creates a tool function for adding or updating a relationship of this class.

        Args:
            add_or_update_relationship_func (function): The function to call to add/update the relationship.

        Returns:
            function: A tool function that can be used by an agent.
        """
        def func(**kwargs):
            logger.system(f"Adding or updating relationship for {self.relationship_name}")
            properties = kwargs.get('kwargs', kwargs)
            
            domain_param_name, range_param_name = self._get_param_names()
            domain_primary_key_value = properties.get(domain_param_name)
            range_primary_key_value = properties.get(range_param_name)
            props = properties.get("properties")
            self._validate_properties(props)
            logger.system(f"Arguments for add_or_update_relationship_func: {self.domain_entity_class}, {self.domain_primary_key_prop}, {domain_primary_key_value}, {self.range_entity_class}, {self.range_primary_key_prop}, {range_primary_key_value}, {self.relationship_name}, {props}, {self.symmetric}")
            return add_or_update_relationship_func(
                self.domain_entity_class,
                self.domain_primary_key_prop,
                domain_primary_key_value, 
                self.range_entity_class,  
                self.range_primary_key_prop,
                range_primary_key_value, 
                self.relationship_name, 
                props, 
                self.symmetric
            )

        func.__name__ = f"add_link_{self.domain_entity_class}_{self.relationship_name}_{self.range_entity_class}"
        func.__doc__ = f"Add or update a [{self.relationship_name}] relationship between a [{self.domain_entity_class}] and [{self.range_entity_class}]\n"+\
            f"Domain Primary Key: from_{self.domain_entity_class}_{self.domain_primary_key_prop}\n"+\
            f"Range Primary Key: to_{self.range_entity_class}_{self.range_primary_key_prop}"+\
            (f"Properties: {self.properties}" if self.properties else "")
        func.__parameters__ = self.get_tool_parameters_schema()
        return func

    def _get_param_names(self):
        """
        Gets the parameter names for the domain and range of the relationship tool.
        Handles the case of self-referential relationships to avoid name collisions.

        Returns:
            tuple: A tuple containing the domain and range parameter names.
        """
        domain_param_name = f"from_{self.domain_entity_class}_{self.domain_primary_key_prop}"
        range_param_name = f"to_{self.range_entity_class}_{self.range_primary_key_prop}"
        return domain_param_name, range_param_name

    def get_tool_parameters_schema(self):
        """
        Builds the JSON schema for the parameters of the relationship tool.

        Returns:
            dict: A dictionary representing the JSON schema.
        """
        domain_param_name, range_param_name = self._get_param_names()

        schema = {
            "type": "object",
            "properties": {
                domain_param_name: {
                    "type": "string",
                    "description": f"The {self.domain_primary_key_prop} of the FROM entity ({self.domain_entity_class})"
                },
                range_param_name: {
                    "type": "string",
                    "description": f"The {self.range_primary_key_prop} of the TO entity ({self.range_entity_class})"
                }
            },
            "required": [domain_param_name, range_param_name]
        }

        if self.properties:
            props_schema = {
                "type": "object",
                "properties": {},
                "required": []
            }
            for prop in self.properties:
                prop_type = "string"
                if prop.type == "float":
                    prop_type = "number"
                elif prop.type == "integer":
                    prop_type = "integer"

                props_schema["properties"][prop.property_name] = {
                    "type": prop_type,
                    "description": prop.description
                }
                props_schema["required"].append(prop.property_name)
            
            schema["properties"]["properties"] = props_schema
        
        return schema

    def get_tool_get_relationship_properties(self, get_relationship_properties_func):
        """
        Creates a tool for getting the properties of a specific relationship instance.

        Args:
            get_relationship_properties_func (function): The function to call to get relationship properties.

        Returns:
            function: A tool function that can be used by an agent.
        """
        def func(**kwargs):
            logger.system(f"Getting relationship properties for {self.relationship_name}")
            properties = kwargs.get('kwargs', kwargs)
            domain_param_name, range_param_name = self._get_param_names()
            domain_primary_key_value = properties.get(domain_param_name)
            range_primary_key_value = properties.get(range_param_name)
            logger.system(f"Arguments for get_relationship_properties_func: {self.domain_entity_class}, {self.domain_primary_key_prop}, {domain_primary_key_value}, {self.relationship_name}, {self.range_entity_class}, {self.range_primary_key_prop}, {range_primary_key_value}")
            return get_relationship_properties_func( self.domain_entity_class, self.domain_primary_key_prop, domain_primary_key_value, self.relationship_name,self.range_entity_class, self.range_primary_key_prop, range_primary_key_value)

        domain_param_name, range_param_name = self._get_param_names()


        func.__name__ = f"get_{self.relationship_name}_properties"
        func.__doc__ = f"Get a {self.relationship_name} relationship properties between _{self.domain_entity_class}_{self.range_entity_class}.\n"+\
            f"Domain Primary Key: from_{self.domain_entity_class}_{self.domain_primary_key_prop}\n"+\
            f"Range Primary Key: to_{self.range_entity_class}_{self.range_primary_key_prop}"+\
            (f"Returns properties of the relationship: {self.properties}" if self.properties else "")
        
            
        func.__parameters__ = {
            "type": "object",
            "properties": {
                domain_param_name: {"type": "string", "description": f"The {self.domain_primary_key_prop} of the FROM entity ({self.domain_entity_class})"}, 
                range_param_name: {"type": "string", "description": f"The {self.range_primary_key_prop} of the TO entity ({self.range_entity_class})"}
            },
            "required": [domain_param_name, range_param_name]
        }
        return func

    def get_tool_get_relationship_entities(self, get_relationship_entities_func):
        """
        Creates a tool for getting all entities connected by a specific relationship.

        Args:
            get_relationship_entities_func (function): The function to call to get the related entities.

        Returns:
            function: A tool function that can be used by an agent.
        """
        def func(**kwargs):
            logger.system(f"Getting relationship entities for {self.relationship_name}")
            properties = kwargs.get('kwargs', kwargs)
            domain_param_name, _ = self._get_param_names()
            domain_primary_key_value = properties.get(domain_param_name)
            logger.system(f"Arguments for get_relationship_entities_func: {self.domain_entity_class}, {self.domain_primary_key_prop}, {domain_primary_key_value}, {self.relationship_name}, {self.range_entity_class}, {self.range_primary_key_prop}")
            return get_relationship_entities_func( self.domain_entity_class, self.domain_primary_key_prop, domain_primary_key_value, self.relationship_name, self.range_entity_class, self.range_primary_key_prop)

        func.__name__ = f"get_{self.range_entity_class}s_{self.domain_entity_class}_{self.relationship_name}"
        func.__doc__ = f"Get all {self.range_entity_class}s linked to a {self.domain_entity_class} in a {self.relationship_name} relationship.\n"+\
            f"Domain Primary Key: from_{self.domain_entity_class}_{self.domain_primary_key_prop}\n"+\
                "Returns a list of {self.range_entity_class}s"
 
        domain_param_name, _ = self._get_param_names()

        func.__parameters__ = {
            "type": "object",
            "properties": {
                domain_param_name: {"type": "string", "description": f"The {self.domain_primary_key_prop} of the FROM entity ({self.domain_entity_class})"}
            },
            "required": [domain_param_name]
        }
        return func



import asyncio
import inspect
import sys
from a1facts.enrichment.knowledge_source import KnowledgeSource
from mcp import ClientSession, StdioServerParameters, types
from mcp.client.stdio import stdio_client
from agno.agent import Agent
from textwrap import dedent
from a1facts.utils.modelconfig import my_model
from datetime import date
from threading import Thread
import atexit

class MCPKnowledgeSource(KnowledgeSource):
    def __init__(self, source_config: dict):
        self.name = source_config['name']
        self.description = source_config['description']
        self.override_reliability = source_config.get('override_reliability')
        self.override_credibility = source_config.get('override_credibility')
        self.session = None
        self.local_tools = []
        self.query_agent = None
        self.stdio_cm = None

        self.loop = asyncio.new_event_loop()
        self.thread = Thread(target=self.loop.run_forever, daemon=True)
        self.thread.start()

        server_params = StdioServerParameters(
            command=source_config['run_command'],
            args=source_config['run_args'],
            cwd=source_config['run_cwd'],
            env=None,
        )

        if 'all_tools' in source_config and source_config['all_tools']:
            self.all_tools = True
        else:
            self.all_tools = False
        if self.all_tools:
            self.desired_tools = None
        else:
            self.desired_tools = source_config['tools']
        
        future = asyncio.run_coroutine_threadsafe(self._init_mcp(server_params), self.loop)
        future.result()

    def __del__(self):
        self.close()

    def close(self):
        if self.session:
            future = asyncio.run_coroutine_threadsafe(self.session.close(), self.loop)
            future.result(timeout=1)
            self.session = None
        if self.stdio_cm:
            future = asyncio.run_coroutine_threadsafe(self.stdio_cm.__aexit__(None, None, None), self.loop)
            future.result(timeout=1)
            self.stdio_cm = None
        
        if self.loop.is_running():
            self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join(timeout=1)

    async def _init_mcp(self, server_params):
        """Initializes the MCP session and creates local tools."""
        self.session = await self._initialize_session(server_params)
        self.local_tools = await self._create_local_tools()
        self.query_agent = self._create_query_agent()
        if sys.platform == "win32":
            await asyncio.sleep(0.2)

    async def _initialize_session(self, server_params):
        self.stdio_cm = stdio_client(server_params)
        read, write = await self.stdio_cm.__aenter__()
        session = ClientSession(read, write, sampling_callback=None)
        await session.initialize()
        # We don't close the session here; it will be kept alive for the tools.
        # A __del__ method could be added to properly close it when the object is destroyed.
        return session

    async def _create_local_tools(self):
        remote_tools_response = await self.session.list_tools()
        if self.desired_tools: #TODO if all_tools and desired set, throw error
            tasks = [self._create_local_tool(tool) for tool in remote_tools_response.tools if tool.name in self.desired_tools]
        else: #create all tools
            tasks = [self._create_local_tool(tool) for tool in remote_tools_response.tools]
        local_tools = await asyncio.gather(*tasks)
        return local_tools
    
    def _create_query_agent(self):
        return Agent(
            name=f"{self.name} Query Agent",
            role=f"Query the {self.name} knowledge source.",
            model=my_model,
            tools=self.local_tools,
            instructions=dedent(f"""
                Query the knowledge source for the information requested by the user.
                Today's date is {date.today().strftime('%Y-%m-%d')}.
            """),
            markdown=True,
            debug_mode=False,
        )

    async def _create_local_tool(self, tool_details: types.Tool):
        """Dynamically creates a local, async function that calls a remote MCP tool."""
        
        async def tool_wrapper(*args, **kwargs):
            param_names = list(tool_details.inputSchema.get('properties', {}).keys())
            final_kwargs = {}

            if 'args' in kwargs or 'kwargs' in kwargs:
                agent_kwargs = kwargs.get('kwargs', {})
                if isinstance(agent_kwargs, dict): final_kwargs.update(agent_kwargs)
                agent_args = kwargs.get('args', [])
                if isinstance(agent_args, dict): final_kwargs.update(agent_args)
                elif isinstance(agent_args, list):
                    for i, arg_value in enumerate(agent_args):
                        if i < len(param_names): final_kwargs[param_names[i]] = arg_value
            else:
                for i, arg_value in enumerate(args):
                    if i < len(param_names): final_kwargs[param_names[i]] = arg_value
                final_kwargs.update(kwargs)
            
            filtered_kwargs = {k: v for k, v in final_kwargs.items() if k in param_names}

            print(f"Calling remote tool: {tool_details.name} with args: {filtered_kwargs}")
            try:
                result = await self.session.call_tool(tool_details.name, arguments=filtered_kwargs)
                if result.isError or not result.content:
                    error_message = result.content[0].text if result.content and isinstance(result.content[0], types.TextContent) else "Unknown error"
                    return f"Error executing tool {tool_details.name}: {error_message}"
                return result.content[0].text
            except Exception as e:
                return f"An exception occurred while calling tool {tool_details.name}: {e}"

        tool_wrapper.__name__ = tool_details.name
        doc = f"{tool_details.description or 'No description available.'}\n\nArgs:\n"
        if 'properties' in tool_details.inputSchema:
            for arg_name, arg_details in tool_details.inputSchema['properties'].items():
                is_required_str = "required" if arg_name in tool_details.inputSchema.get('required', []) else "optional"
                doc += f"    {arg_name} ({arg_details.get('type', 'unknown')}, {is_required_str}): {arg_details.get('description', '')}\n"
        tool_wrapper.__doc__ = doc
        return tool_wrapper

    def query_tool(self):
        async def query_handler(query_text: str) -> str:
            result = await self.query_agent.arun(query_text)
            return result.content
        
        query_handler.__name__ = f"query_{self.name}"
        query_handler.__doc__ = f"Query the {self.name} knowledge source. {self.description}"
        return query_handler

    def __str__(self) -> str:
        return f"{self.name} - {self.description}"




from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.utils.modelconfig import my_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime

class UpdateAgent:
    def __init__(self, ontology: KnowledgeOntology, mytools: list):
        self.ontology = ontology
        self.update_agent = Agent(
            name="Knowledge Graph Update Agent",
            role="Update the knowledge graph",
            model=my_model,
            tools=mytools,

            instructions=dedent(f"""
                The user is providing you unstrucutred knowledge. Translate the knowledge into a structured format based on the ontology.
                Ontology:[{self.ontology}]
                Return the results in RDFS format.
                When you are done, add every entity and relationship to the graph using the tools available to you.
                First add the entities, then add the relationships.
                Make sure to add every single one of them.
                Today is {datetime.now().strftime("%Y-%m-%d")}
            """),
            markdown=True,
            debug_mode=False,
            )

    def update(self, knowledge: str):
        return self.update_agent.run("Translate the following knowledge into a structured format based on the ontology, then add every entity and every relationship to the graph using the tools available to you.\n \n " + knowledge)



import importlib
from a1facts.enrichment.function_knowledge_source.query_agent import QueryAgent
from a1facts.enrichment.knowledge_source import KnowledgeSource
from a1facts.utils.logger import logger

class FunctionKnowledgeSource(KnowledgeSource):
    def __init__(self, source_config: dict):
        self.name = source_config['name']
        self.description = source_config['description']
        self.functions_package = source_config['functions_package']
        self.override_reliability = source_config['override_reliability']
        self.override_credibility = source_config['override_credibility']
        self.tools = []
        self.query_agent = None
        logger.system(f"Initializing FunctionKnowledgeSource for {self.name}")
        self._validate_source_config(source_config)


    def _validate_source_config(self, source_config: dict):
        if 'name' not in source_config:
            raise ValueError("Your knowledge source config is missing the 'name' field")
        if 'description' not in source_config:
            raise ValueError("Your knowledge source config is missing the 'description' field")
        if 'functions_package' not in source_config:
            raise ValueError("Your knowledge source config is missing the 'functions_file' field")
        if 'override_reliability' not in source_config:
            raise ValueError("Your knowledge source config is missing the 'override_reliability' field")
        if 'override_credibility' not in source_config:
            raise ValueError("Your knowledge source config is missing the 'override_credibility' field")

    def query_tool(self):
        functions_module = importlib.import_module(f"{self.functions_package}")
        for func_name in dir(functions_module):
            logger.system(f"Checking function {func_name}")
            if func_name.startswith('__') and func_name.endswith('__'):
                continue
            if hasattr(getattr(functions_module, func_name), '__class__') and func_name[0].isupper():
                continue
            if func_name == "load_dotenv":
                continue
            if not func_name.startswith('_'):                
                func = getattr(functions_module, func_name)
                if callable(func):
                    logger.system(f"Adding function {func_name} to tools")
                    self.tools.append(func)
        logger.system(f"Tools for {self.name}: {self.tools}")
        self.query_agent = QueryAgent(self.tools)
        logger.system(f"Query agent for {self.name} initialized")
        def query_handler(query_text):
            return self.query_agent.query(query_text)

        query_handler.__name__ = self.name+"_query"
        query_handler.__doc__ = self.description + \
            "ARGS: query_text: str - The query text to query the knowledge source\n" + \
            "RETURNS: str - The response from the knowledge source\n" + \
            "Override reliability: " + self.override_reliability + "\n" + \
            "Override credibility: " + self.override_credibility
        return query_handler

    def __str__(self) -> str:
        return self.name + " - " + self.description + " - " + self.override_reliability + " - " + self.override_credibility + " - " + self.functions_file

